from django.db import migrations, models

from comment.utils import id_generator


def generate_urlhash():
    return id_generator(
        prefix='comment',
        len_id=8,
        suffix=''
        )


def set_unique_urlhash(model, instance):
    if not instance.urlhash:
        instance.urlhash = generate_urlhash()
        while model.objects.filter(urlhash=instance.urlhash).exists():
            instance.urlhash = generate_urlhash()


def set_default_urlhash(apps, schema_editor):
    comment_model = apps.get_model('comment', 'Comment')
    for comment in comment_model.objects.all():
        set_unique_urlhash(comment_model, comment)
        comment.save(update_fields=['urlhash'])


class Migration(migrations.Migration):

    dependencies = [
        ('comment', '0007_auto_20200620_1259'),
    ]

    operations = [
        migrations.AddField(
            model_name='comment',
            name='urlhash',
            field=models.CharField(null=True, default=None, editable=False, max_length=50),
        ),
        migrations.RunPython(set_default_urlhash, migrations.RunPython.noop),
        migrations.AlterField(
            model_name='comment',
            name='urlhash',
            field=models.CharField(editable=False, max_length=50, unique=True)
        )
    ]



from django.dispatch import receiver
from django.db.models import signals

from comment.models import FlagInstance, ReactionInstance


@receiver(signals.post_delete, sender=FlagInstance)
def decrease_count(sender, instance, using, **kwargs):
    """Decrease flag count in the flag model before deleting an instance"""
    instance.flag.decrease_count()
    instance.flag.toggle_flagged_state()


@receiver(signals.post_delete, sender=ReactionInstance)
def delete_reaction_instance(sender, instance, using, **kwargs):
    instance.reaction.decrease_reaction_count(instance.reaction_type)



from django.core.exceptions import ImproperlyConfigured

from comment.conf import settings
from comment.messages import ErrorMessage
from comment.utils import get_request_data, is_comment_moderator, paginate_comments, get_model_obj, has_valid_profile


class DABContext(dict):
    def __init__(self, request, model_object=None, **kwargs):
        self.request = request
        self.app_name = get_request_data(request, 'app_name')
        self.model_name = get_request_data(request, 'model_name')
        self.model_id = get_request_data(request, 'model_id')
        self.model_object = model_object
        if not self.model_object:
            self.model_object = get_model_obj(self.app_name, self.model_name, self.model_id)
        super().__init__(**self(), **kwargs)

    @staticmethod
    def get_login_url():
        login_url = settings.LOGIN_URL
        if not login_url:
            raise ImproperlyConfigured(ErrorMessage.LOGIN_URL_MISSING)
        if not login_url.startswith('/'):
            login_url = '/' + login_url
        return login_url

    def is_oauth(self):
        oauth = get_request_data(self.request, 'oauth')
        if oauth and oauth.lower() == 'true':
            return True
        return False

    def get_comments(self):
        comments = self.model_object.comments.filter_parents_by_object(
            self.model_object, include_flagged=is_comment_moderator(self.request.user)
        )
        page = get_request_data(self.request, 'page')
        comments_per_page = settings.COMMENT_PER_PAGE
        if comments_per_page:
            comments = paginate_comments(comments, comments_per_page, page)
        return comments

    def __call__(self):
        return {
            'model_object': self.model_object,
            'model_name': self.model_name,
            'model_id': self.model_id,
            'app_name': self.app_name,
            'user': self.request.user,
            'comments': self.get_comments(),
            'login_url': self.get_login_url(),
            'has_valid_profile': has_valid_profile(),
            'allowed_flags': settings.COMMENT_FLAGS_ALLOWED,
            'is_anonymous_allowed': settings.COMMENT_ALLOW_ANONYMOUS,
            'is_translation_allowed': settings.COMMENT_ALLOW_TRANSLATION,
            'is_subscription_allowed': settings.COMMENT_ALLOW_SUBSCRIPTION,
            'oauth': self.is_oauth()
        }



from unittest.mock import patch

from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory
from django.utils import timezone
from django.core import signing, mail
from django.urls import reverse
from rest_framework import status
from django.contrib import messages

from comment.conf import settings
from comment.models import Comment, Follower
from comment.messages import EmailInfo, EmailError
from comment.tests.base import BaseCommentViewTest
from comment.tests.test_utils import BaseAnonymousCommentTest
from comment.views import ConfirmComment


class CommentViewTestCase(BaseCommentViewTest):
    def setUp(self):
        super().setUp()
        self.all_comments = Comment.objects.all().count()
        self.parent_comments = Comment.objects.all_parents().count()
        self.data = {
            'content': 'comment body',
            'app_name': 'post',
            'model_name': 'post',
            'parent_id': '',
            'model_id': self.post_1.id
        }

    def increase_count(self, parent=False):
        if parent:
            self.parent_comments += 1

        self.all_comments += 1

    @staticmethod
    def get_create_url():
        return reverse('comment:create')

    def comment_count_test(self):
        self.assertEqual(Comment.objects.all_parents().count(), self.parent_comments)
        self.assertEqual(Comment.objects.all().count(), self.all_comments)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_create_parent_comment(self):
        self.assertEqual(self.all_comments, 0)
        self.assertEqual(self.parent_comments, 0)

        url = self.get_create_url()
        response = self.client.post(url, data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'comment/comments/base.html')
        self.assertHtmlTranslated(response.content, url)

        parent_comment = Comment.objects.get(object_id=self.post_1.id, parent=None)
        self.assertEqual(response.context.get('comment').id, parent_comment.id)
        self.assertTrue(response.context.get('comment').is_parent)

        self.increase_count(parent=True)
        self.comment_count_test()

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_create_child_comment(self):
        self.assertEqual(self.all_comments, 0)
        self.assertEqual(self.parent_comments, 0)

        url = self.get_create_url()
        data = self.data.copy()
        parent_comment = self.create_comment(self.post_1)
        self.increase_count(parent=True)
        self.comment_count_test()
        data['parent_id'] = parent_comment.id
        response = self.client.post(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'comment/comments/child_comment.html')
        self.assertHtmlTranslated(response.content, url)

        child_comment = Comment.objects.get(object_id=self.post_1.id, parent=parent_comment)
        self.assertEqual(response.context.get('comment').id, child_comment.id)
        self.assertFalse(response.context.get('comment').is_parent)

        self.increase_count()
        self.comment_count_test()

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_send_notification_to_thread_followers_on_create_comment(self):
        Follower.objects.follow('te@te.com', 'test_user', self.post_1)
        self.assertEqual(len(mail.outbox), 0)
        url = self.get_create_url()
        response = self.client.post(url, data=self.data)
        self.assertEqual(response.status_code, 200)
        response.context['view'].email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)

    def test_create_comment_non_ajax_request(self):
        response = self.client_non_ajax.post(self.get_create_url(), data=self.data)
        self.assertEqual(response.status_code, 400)

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_anonymous_comment(self):
        self.client.logout()
        data = self.data.copy()
        data['email'] = 'a@a.com'
        url = self.get_create_url()
        self.assertEqual(len(mail.outbox), 0)

        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTemplateUsed(response, 'comment/comments/base.html')
        self.assertEqual(response.json()['msg'], EmailInfo.CONFIRMATION_SENT)
        # confirmation email is sent
        response.context['view'].email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)

        # no change in comment count
        self.comment_count_test()

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_anonymous_comment_with_invalid_email(self):
        self.client.logout()
        data = self.data.copy()
        data['email'] = 'test@invalid.c'
        url = self.get_create_url()

        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()['error'], EmailError.EMAIL_INVALID)


class TestEditComment(BaseCommentViewTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment = cls.create_comment(cls.content_object_1)
        cls.init_content = cls.comment.content

    def test_edit_comment(self):
        comment = self.create_comment(self.content_object_1)
        self.client.force_login(comment.user)
        self.assertEqual(Comment.objects.all().count(), 2)
        data = {
            'content': 'parent comment was edited',
            'app_name': 'post',
            'model_name': 'post',
            'model_id': self.post_1.id
        }
        get_url = self.get_url('comment:edit', comment.id, data)
        self.assertEqual(comment.content, 'comment 2')
        response = self.client.get(get_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed('comment/comments/update_comment.html')
        self.assertHtmlTranslated(response.content, get_url)
        self.assertEqual(response.context['comment_form'].instance.id, comment.id)

        post_url = self.get_url('comment:edit', comment.id)
        response = self.client.post(post_url, data=data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed('comment/comments/comment_content.html')
        self.assertHtmlTranslated(response.content, post_url)
        comment.refresh_from_db()
        self.assertEqual(comment.content, data['content'])

        data['content'] = ''
        with self.assertRaises(ValueError) as error:
            self.client.post(
                self.get_url('comment:edit', comment.id), data=data, HTTP_X_REQUESTED_WITH='XMLHttpRequest'
            )
        self.assertIsInstance(error.exception, ValueError)

    def test_cannot_edit_comment_by_different_user(self):
        comment = self.comment
        self.client.force_login(self.user_2)
        data = {
            'content': 'parent comment was edited',
            'app_name': 'post',
            'model_name': 'post',
            'model_id': self.post_1.id
        }
        self.assertEqual(comment.user.username, self.user_1.username)
        response = self.client.get(self.get_url('comment:edit', comment.id), data=data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.reason_phrase, 'Forbidden')

        response = self.client.post(self.get_url('comment:edit', comment.id), data=data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.reason_phrase, 'Forbidden')


class TestDeleteComment(BaseCommentViewTest):

    def response_fails_test(self, response):
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.reason_phrase, 'Forbidden')

    def test_delete_comment(self):
        comment = self.create_comment(self.content_object_1)
        self.client.force_login(comment.user)
        init_count = Comment.objects.all().count()
        self.assertEqual(init_count, 1)
        get_url = self.get_url('comment:delete', comment.id, self.data)
        response = self.client.get(get_url, data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'comment/comments/comment_modal.html')
        self.assertContains(response, 'data')
        self.assertHtmlTranslated(response.json()['data'], get_url)

        post_url = self.get_url('comment:delete', comment.id)
        response = self.client.post(post_url, data=self.data)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'comment/comments/base.html')
        self.assertHtmlTranslated(response.content, post_url)
        self.assertNotContains(response, 'html_form')
        self.assertRaises(Comment.DoesNotExist, Comment.objects.get, id=comment.id)
        self.assertEqual(Comment.objects.all().count(), init_count-1)

    def test_delete_comment_by_moderator(self):
        comment = self.create_comment(self.content_object_1)
        self.client.force_login(self.moderator)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.moderator.id)
        self.assertTrue(self.moderator.has_perm('comment.delete_flagged_comment'))
        self.assertEqual(comment.user, self.user_1)
        init_count = Comment.objects.count()
        self.assertEqual(init_count, 1)
        # moderator cannot delete un-flagged comment
        response = self.client.post(self.get_url('comment:delete', comment.id), data=self.data)
        self.assertEqual(response.status_code, 403)

        # moderator can delete flagged comment
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1):
            self.create_flag_instance(self.user_1, comment)
            self.create_flag_instance(self.user_2, comment)
            response = self.client.post(self.get_url('comment:delete', comment.id), data=self.data)
            self.assertEqual(response.status_code, 200)
            self.assertRaises(Comment.DoesNotExist, Comment.objects.get, id=comment.id)

    def test_delete_comment_by_admin(self):
        comment = self.create_comment(self.content_object_1)
        self.client.force_login(self.admin)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.admin.id)
        self.assertTrue(self.admin.groups.filter(name='comment_admin').exists())
        self.assertEqual(comment.user, self.user_1)
        init_count = Comment.objects.count()
        self.assertEqual(init_count, 1)

        # admin can delete any comment
        response = self.client.post(self.get_url('comment:delete', comment.id), data=self.data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Comment.objects.count(), init_count - 1)

    def test_cannot_delete_comment_by_different_user(self):
        comment = self.create_comment(self.content_object_1)
        self.client.force_login(self.user_2)
        self.assertEqual(comment.content, 'comment 1')
        self.assertEqual(comment.user.username, self.user_1.username)

        init_count = Comment.objects.all().count()
        self.assertEqual(init_count, 1)

        # test GET request
        response = self.client.get(self.get_url('comment:delete', comment.id), data=self.data)
        self.response_fails_test(response)

        # test POST request
        response = self.client.post(self.get_url('comment:delete', comment.id), data=self.data)
        self.response_fails_test(response)


class ConfirmCommentViewTest(BaseAnonymousCommentTest):
    def setUp(self):
        super().setUp()
        # although this will work for authenticated users as well.
        self.client.logout()
        self.request.user = AnonymousUser()
        self.init_count = Comment.objects.all().count()
        self.template_1 = 'comment/anonymous/discarded.html'
        self.template_2 = 'comment/comments/messages.html'
        self.template_3 = 'comment/bootstrap.html'

    def get_url(self, key=None):
        if not key:
            key = self.key
        return reverse('comment:confirm-comment', args=[key])

    def template_used_test(self, response):
        self.assertTemplateUsed(response, self.template_1)
        self.assertTemplateUsed(response, self.template_2)
        self.assertTemplateUsed(response, self.template_3)

    def test_bad_signature(self):
        key = self.key + 'invalid'
        _url = self.get_url(key)
        response = self.client.get(_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(Comment.objects.all().count(), self.init_count)
        self.template_used_test(response)
        response_messages = response.context['messages']
        self.assertEqual(len(response_messages), 1)
        for r in response_messages:
            self.assertEqual(r.level, messages.ERROR)
            self.assertEqual(r.message, EmailError.BROKEN_VERIFICATION_LINK)
            self.assertTextTranslated(r.message, _url)

    def test_comment_exists(self):
        comment_dict = self.comment_obj.to_dict().copy()
        comment = self.create_anonymous_comment(posted=timezone.now(), email='a@a.com')
        init_count = self.init_count + 1
        comment_dict.update({
            'posted': str(comment.posted),
            'email': comment.email
        })
        key = signing.dumps(comment_dict)
        _url = self.get_url(key)
        response = self.client.get(_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(Comment.objects.all().count(), init_count)
        self.template_used_test(response)
        self.assertHtmlTranslated(response.content, _url)
        response_messages = response.context['messages']
        self.assertEqual(len(response_messages), 1)
        for r in response_messages:
            self.assertEqual(r.level, messages.WARNING)
            self.assertEqual(r.message, EmailError.USED_VERIFICATION_LINK)
            self.assertTextTranslated(r.message, _url)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_success_without_notification(self):
        response = self.client.get(self.get_url())
        comment = Comment.objects.get(email=self.comment_obj.email, posted=self.time_posted)
        self.assertEqual(Comment.objects.all().count(), self.init_count + 1)
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, comment.get_url(self.request))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_success_with_notification(self):
        self.assertEqual(len(mail.outbox), 0)
        Follower.objects.follow('t@r.com', 'test_user', self.post_1)
        request = RequestFactory().get(self.get_url())
        request.user = self.user_2
        view = ConfirmComment()
        response = view.get(request, key=self.key)
        comment = Comment.objects.get(email=self.comment_obj.email, posted=self.time_posted)
        self.assertEqual(Comment.objects.all().count(), self.init_count + 1)
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, comment.get_url(self.request))

        view.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)



from urllib.parse import quote_plus
from unittest.mock import patch

from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.test import TestCase, RequestFactory, TransactionTestCase, Client
from django.urls import reverse, resolve
from django.utils import timezone, translation
from lxml.html.soupparser import fromstring
from lxml.cssselect import CSSSelector

from comment.conf import settings
from comment.models import Comment, FlagInstance, Reaction, ReactionInstance
from post.models import Post


User = get_user_model()


@patch.object(settings, 'COMMENT_ALLOW_TRANSLATION', True)
class BaseInternationalizationTest:
    translatable_attrs = ['title', 'aria-label']

    @staticmethod
    def has_translatable_html_text(element):
        if "comment-translatable" in element.attrib.get('class', '').split():
            return True
        return False

    @staticmethod
    def has_translatable_html_attr(element, attr):
        if element.attrib.get(attr, '').strip():
            return True
        return False

    @staticmethod
    def is_translated(text):
        try:
            text.encode('ascii')
        except UnicodeEncodeError:
            return True
        return False

    @staticmethod
    def get_view_from_url_or_none(url):
        if url:
            clean_url = url.split('?')[0]
            return resolve(clean_url)[0].__name__


class BaseCommentTest(TestCase, BaseInternationalizationTest):
    flags = 0
    reactions = 0
    content_object_1 = None
    increment = 0
    user_1 = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user_1 = User.objects.create_user(
                    username="test-1",
                    email="test-1@acme.edu",
                    password="1234"
        )
        cls.user_2 = User.objects.create_user(
            username="test-2",
            email="test-2@acme.edu",
            password="1234"
        )
        cls.user_without_email = User.objects.create_user(
            username="no-email",
            password="1234"
        )
        cls.moderator = User.objects.create_user(
            username="moderator",
            email="test-2@acme.edu",
            password="1234"
        )
        moderator_group = Group.objects.filter(name='comment_moderator').first()
        moderator_group.user_set.add(cls.moderator)
        cls.admin = User.objects.create_user(
            username="admin",
            email="test-2@acme.edu",
            password="1234"
        )
        admin_group = Group.objects.filter(name='comment_admin').first()
        admin_group.user_set.add(cls.admin)
        cls.post_1 = Post.objects.create(
            author=cls.user_1,
            title="post 1",
            body="first post body"
        )
        cls.post_2 = Post.objects.create(
            author=cls.user_1,
            title="post 2",
            body="second post body"
        )
        content_type = ContentType.objects.get(model='post')
        cls.content_object_1 = content_type.get_object_for_this_type(id=cls.post_1.id)
        cls.content_object_2 = content_type.get_object_for_this_type(id=cls.post_2.id)
        cls.increment = 0
        cls.reactions = 0
        cls.flags = 0

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        translation.activate("test")
        self.addCleanup(patch.stopall)

    @classmethod
    def increase_comment_count(cls):
        cls.increment += 1

    @classmethod
    def create_comment(cls, ct_object, user=None, email=None, posted=None, parent=None):
        if not user:
            user = cls.user_1
        cls.increase_comment_count()
        return Comment.objects.create(
            content_object=ct_object,
            content='comment {}'.format(cls.increment),
            user=user,
            parent=parent,
        )

    @classmethod
    def create_anonymous_comment(cls, ct_object=None, email=None, posted=None, parent=None):
        if not ct_object:
            ct_object = cls.content_object_1
        if not email:
            email = cls.user_1.email
        if not posted:
            posted = timezone.now()
        cls.increase_comment_count()
        return Comment.objects.create(
            content_object=ct_object,
            content='anonymous comment {}'.format(cls.increment),
            parent=parent,
            email=email,
            posted=posted
        )

    @classmethod
    def create_reaction_instance(cls, user, comment, reaction):
        reaction_type = getattr(ReactionInstance.ReactionType, reaction.upper(), None)
        if reaction_type:
            reaction_obj = Reaction.objects.get(comment=comment)
            cls.reactions += 1
            reaction_instance = ReactionInstance.objects.create(
                user=user,
                reaction_type=reaction_type.value,
                reaction=reaction_obj
            )
            comment.reaction.refresh_from_db()
            return reaction_instance
        raise ValueError('{} is not a valid reaction type'.format(reaction))

    @staticmethod
    def set_reaction(user, comment, reaction):
        ReactionInstance.objects.set_reaction(user, comment.reaction, reaction)

    @staticmethod
    def set_flag(user, comment, **kwargs):
        return FlagInstance.objects.set_flag(user, comment.flag, **kwargs)

    @classmethod
    def create_flag_instance(cls, user, comment, **kwargs):
        instance = FlagInstance.objects.create(
            user=user,
            flag=comment.flag,
            **kwargs
        )
        cls.flags += 1
        return instance

    def _check_translatable_html_text(self, element, url):
        if self.has_translatable_html_text(element):
            self.assertTrue(
                self.is_translated(element.text),
                (f'No translation for the element {element.tag} with text "{element.text}" '
                 f'from view {self.get_view_from_url_or_none(url)}')
            )

    def _check_translatable_html_attrs(self, element, url):
        for attr in self.translatable_attrs:
            if self.has_translatable_html_attr(element, attr):
                self.assertTrue(
                    self.is_translated(element.attrib.get(attr)),
                    (f'No translation for the attribute "{attr}" of the element {element.tag} with the '
                     f'value "{element.attrib.get(attr)}" from view {self.get_view_from_url_or_none(url)}')
                )

    def assertHtmlTranslated(self, html, url=None):
        root = fromstring(html)
        sel = CSSSelector('*')

        for element in sel(root):
            self._check_translatable_html_text(element, url)
            self._check_translatable_html_attrs(element, url)

    def assertTextTranslated(self, text, url=None):
        self.assertTrue(
            self.is_translated(text),
            f'No translation for the text "{text}" from view {self.get_view_from_url_or_none(url)}'
        )

    def assertQuerysetEqual(self, qs, values, transform=None, ordered=True, msg=None):
        if not transform:
            def transform(x):
                return x
        return super().assertQuerysetEqual(qs, values, transform=transform, ordered=True, msg=msg)


class BaseCommentManagerTest(BaseCommentTest):
    content_object_2 = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.parent_comment_1 = cls.create_comment(cls.content_object_1)
        cls.parent_comment_2 = cls.create_comment(cls.content_object_1)
        cls.parent_comment_3 = cls.create_comment(cls.content_object_1)
        cls.child_comment_1 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_1)
        cls.child_comment_2 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_2)
        cls.child_comment_3 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_2)

        cls.parent_comment_4 = cls.create_comment(cls.content_object_2)
        cls.parent_comment_5 = cls.create_comment(cls.content_object_2)
        cls.child_comment_4 = cls.create_comment(cls.content_object_2, parent=cls.parent_comment_1)
        cls.child_comment_5 = cls.create_comment(cls.content_object_2, parent=cls.parent_comment_2)


class BaseCommentViewTest(BaseCommentTest):

    def setUp(self):
        super().setUp()
        self.client = Client(HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.client_non_ajax = Client()
        self.client.force_login(self.user_2)
        self.client_non_ajax.force_login(self.user_2)
        self.data = {
            'content': 'parent comment was edited',
            'app_name': 'post',
            'model_name': 'post',
            'model_id': self.post_1.id,
        }

    @staticmethod
    def get_url(reverse_name, pk=None, data=None):
        if pk:
            url = reverse(reverse_name, args=[pk])
        else:
            url = reverse(reverse_name)

        if not data:
            data = {}

        query_string = '&'.join([f'{name}={quote_plus(str(value))}' for name, value in data.items()])
        if query_string:
            return url + f'?{query_string}'
        return url


class BaseCommentFlagTest(BaseCommentViewTest):
    user_2 = None
    content_object_2 = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment = cls.create_comment(cls.content_object_1)
        cls.comment_for_change_state = cls.create_comment(cls.content_object_1)
        cls.user = cls.user_1
        cls.flag_data = {
            'reason': str(FlagInstance.objects.reason_values[0]),
            'info': None,
        }
        cls.comment_2 = cls.create_comment(cls.content_object_2)
        cls.flag_instance = cls.create_flag_instance(cls.user_2, cls.comment_2, **cls.flag_data)


class BaseTemplateTagsTest(BaseCommentTest):
    class MockUser:
        """Mock unauthenticated user for template. The User instance always returns True for `is_authenticated`"""
        is_authenticated = False

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.factory = RequestFactory()
        settings.PROFILE_APP_NAME = 'user_profile'
        settings.COMMENT_ALLOW_ANONYMOUS = True
        cls.increment = 0
        cls.parent_comment_1 = cls.create_comment(cls.content_object_1)
        cls.parent_comment_2 = cls.create_comment(cls.content_object_1)
        cls.parent_comment_3 = cls.create_comment(cls.content_object_1)
        cls.child_comment_1 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_1)
        cls.child_comment_2 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_2)
        cls.child_comment_3 = cls.create_comment(cls.content_object_1, parent=cls.parent_comment_2)
        cls.anonymous_parent_comment = cls.create_anonymous_comment()
        cls.anonymous_child_comment = cls.create_anonymous_comment(parent=cls.parent_comment_1)


class BaseCommentUtilsTest(BaseCommentTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.factory = RequestFactory()
        cls.comment_1 = cls.create_comment(cls.content_object_1)
        cls.comment_2 = cls.create_comment(cls.content_object_1)
        cls.comment_3 = cls.create_comment(cls.content_object_1)
        cls.anonymous_comment = cls.create_anonymous_comment()


class BaseCommentMigrationTest(TransactionTestCase):
    """
    Test specific migrations
        Make sure that `self.migrate_from` and `self.migrate_to` are defined.
    """

    @property
    def app(self):
        return apps.get_containing_app_config(type(self).__module__).name

    migrate_from = None
    migrate_to = None

    def setUp(self):
        super().setUp()
        assert self.migrate_to and self.migrate_from, \
            f'TestCase {type(self).__name} must define migrate_to and migrate_from properties'
        self.migrate_from = [(self.app, self.migrate_from)]
        self.migrate_to = [(self.app, self.migrate_to)]
        self.executor = MigrationExecutor(connection)
        self.old_apps = self.executor.loader.project_state(self.migrate_from).apps

        self.user = User.objects.create_user(username="tester-1")
        self.post = Post.objects.create(
            author=self.user,
            title="post 3",
            body="third post body"
        )
        content_type = ContentType.objects.get(model='post')
        self.ct_object = content_type.get_object_for_this_type(id=self.post.id)

        # revert to the original migration
        self.executor.migrate(self.migrate_from)
        # ensure return to the latest migration, even if the test fails
        self.addCleanup(self.force_migrate)

        self.setUpBeforeMigration(self.old_apps)
        self.executor.loader.build_graph()
        self.executor.migrate(self.migrate_to)
        self.new_apps = self.executor.loader.project_state(self.migrate_to).apps

    def setUpBeforeMigration(self, apps):
        pass

    @property
    def new_model(self):
        return self.new_apps.get_model(self.app, 'Comment')

    @property
    def old_model(self):
        return self.old_apps.get_model(self.app, 'Comment')

    def force_migrate(self, migrate_to=None):
        self.executor.loader.build_graph()  # reload.
        if migrate_to is None:
            # get latest migration of current app
            migrate_to = [key for key in self.executor.loader.graph.leaf_nodes() if key[0] == self.app]
        self.executor.migrate(migrate_to)


class BaseCommentMixinTest(BaseCommentTest):
    base_url = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.factory = RequestFactory()
        cls.data = {
            'content': 'test',
            'model_name': 'post',
            'app_name': 'post',
            'model_id': 1
        }
        cls.comment = cls.create_comment(cls.post_1, cls.user_1)

    def get_url(self, base_url=None, **kwargs):
        if not base_url:
            base_url = self.base_url
        if kwargs:
            base_url += '?'
            for (key, val) in kwargs.items():
                base_url += str(key) + '=' + str(val) + '&'
        return base_url.rstrip('&')


class BaseCommentSignalTest(BaseCommentManagerTest):
    def setUp(self):
        super().setUp()
        self.user = self.user_1
        self.comment = self.child_comment_1
        self.LIKE = ReactionInstance.ReactionType.LIKE
        self.DISLIKE = ReactionInstance.ReactionType.DISLIKE
        self.flag_data = {
            'reason': str(FlagInstance.objects.reason_values[0]),
            'info': None,
        }



from unittest.mock import patch

from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.shortcuts import reverse

from comment.conf import settings
from comment.tests.test_utils import BaseAnonymousCommentTest
from comment.service.email import DABEmailService
from comment.messages import EmailInfo
from comment.models import Follower


@patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
@patch.object(settings, 'COMMENT_FROM_EMAIL', 'no-reply@domain')
@patch.object(settings, 'COMMENT_CONTACT_EMAIL', 'contact@domain')
class TestDABEmailService(BaseAnonymousCommentTest):
    def setUp(self):
        super().setUp()
        self.email_service = DABEmailService(self.comment_obj, self.request)
        self.confirmation_url = reverse('comment:confirm-comment', args=[self.key])
        self.confirmation_url_drf = f'/api/comments/confirm/{self.key}/'
        self.contact_email = settings.COMMENT_CONTACT_EMAIL
        self.receivers = []
        self.sender = settings.COMMENT_FROM_EMAIL
        self.site = self.email_service.get_msg_context()['site']
        self.content_object_url = f'http://{self.site.domain}{self.comment_obj.content_object.get_absolute_url()}'

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.child_comment = cls.create_comment(cls.content_object_1, parent=cls.comment_1)

    def email_contents_test(self, contents, *args):
        for arg in args:
            self.assertTrue(arg in contents)
        self.assertTrue(self.site.name in contents)
        self.assertTrue(self.content_object_url in contents)

    def email_metadata_test(self, email):
        self.assertEqual(email.from_email, self.sender)
        self.assertEqual(email.to, self.receivers)
        if self.email_service.is_html:
            self.assertEqual(email.alternatives[0][1], 'text/html')
        else:
            self.assertEqual(email.alternatives, [])

    def test_can_create_service(self):
        self.assertIsNotNone(self.email_service)

    def test_get_msg_context(self):
        context = self.email_service.get_msg_context(test='test')
        self.assertEqual(context['comment'], self.comment_obj)
        self.assertEqual(context['contact'], settings.COMMENT_CONTACT_EMAIL)
        self.assertEqual(context['test'], 'test')
        self.assertIn('site', context.keys())

    def test_get_message_for_txt(self):
        subject = 'test subject'
        body = 'test body'
        msg = self.email_service.get_message(subject, body, self.receivers)
        self.assertIsInstance(msg, EmailMultiAlternatives)
        self.assertEqual(msg.to, self.receivers)
        self.assertEqual(msg.alternatives, [])

    def test_get_message_for_html(self):
        subject = 'test subject'
        body = 'test body'
        html = 'html_test'
        msg = self.email_service.get_message(subject, body, self.receivers, html_msg=html)
        self.assertIsInstance(msg, EmailMultiAlternatives)
        self.assertEqual(msg.to, self.receivers)
        self.assertEqual(msg.alternatives, [(html, 'text/html')])

    @patch.object(settings, 'COMMENT_SEND_HTML_EMAIL', False)
    def test_get_message_templates_text_only(self):
        email_service = DABEmailService(self.comment_obj, self.request)
        context = self.email_service.get_msg_context()
        text_template = 'comment/anonymous/confirmation_request.txt'
        html_template = 'comment/anonymous/confirmation_request.html'
        text_msg, html_msg = email_service.get_message_templates(text_template, html_template, context)
        self.assertFalse(email_service.is_html)
        self.assertIsNotNone(text_msg)
        self.assertIsNone(html_msg)

    def test_get_message_templates_with_html(self):
        context = self.email_service.get_msg_context()
        text_template = 'comment/anonymous/confirmation_request.txt'
        html_template = 'comment/anonymous/confirmation_request.html'
        text_msg, html_msg = self.email_service.get_message_templates(text_template, html_template, context)
        self.assertTrue(self.email_service.is_html)
        self.assertIsNotNone(text_msg)
        self.assertIsNotNone(html_msg)

    def test_send_messages(self):
        self.receivers = ['test@test']
        messages = [
            self.email_service.get_message('test subject', 'test body', ['test@test'], html_msg='test msg')
            for _ in range(100)
        ]
        self.assertEqual(len(mail.outbox), 0)
        self.email_service.send_messages(messages)
        self.assertNotEqual(len(mail.outbox), 100)
        self.assertTrue(self.email_service._email_thread.is_alive())
        self.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 100)

    def test_send_confirmation_request_django(self):
        self.email_service.send_confirmation_request()
        self.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)
        sent_email = mail.outbox[0]
        self.assertIsInstance(sent_email, EmailMultiAlternatives)
        self.assertEqual(sent_email.subject, EmailInfo.CONFIRMATION_SUBJECT)
        self.email_contents_test(
            sent_email.body, self.confirmation_url, self.comment_obj.content, self.contact_email
        )
        self.receivers = [self.comment_obj.to_dict()['email']]
        self.email_metadata_test(sent_email)

    def test_send_confirmation_request_api(self):
        self.email_service.send_confirmation_request(api=True)
        self.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)
        sent_email = mail.outbox[0]
        self.assertIsInstance(sent_email, EmailMultiAlternatives)
        self.email_contents_test(
            sent_email.body, self.confirmation_url_drf, self.comment_obj.content, self.contact_email
        )
        self.assertEqual(sent_email.subject, EmailInfo.CONFIRMATION_SUBJECT)
        self.receivers = [self.comment_obj.to_dict()['email']]
        self.email_metadata_test(sent_email)

    def test_get_thread_for_parent_comment(self):
        """The content type (Post, Picture...) is the parent thread of a parent comment"""
        self.assertTrue(self.email_service.comment.is_parent)
        thread = self.email_service.get_thread()
        self.assertIs(thread, self.email_service.comment.content_object)
        self.assertIsInstance(thread, self.email_service.comment.content_object.__class__)

    def test_get_thread_for_child_comment(self):
        """The parent comment is the parent thread of a child comment"""
        self.assertFalse(self.child_comment.is_parent)
        email_service = DABEmailService(self.child_comment, self.request)
        self.assertFalse(email_service.comment.is_parent)
        thread = email_service.get_thread()
        self.assertIsNot(thread, email_service.comment.content_object)
        self.assertIs(thread, email_service.comment.parent)

    def test_get_thread_name_for_parent_comment(self):
        self.assertTrue(self.email_service.comment.is_parent)
        thread_name = self.email_service.get_thread_name()
        self.assertEqual(thread_name, str(self.email_service.comment.content_object))

    def test_get_thread_name_for_child_comment(self):
        email_service = DABEmailService(self.child_comment, self.request)
        self.assertFalse(email_service.comment.is_parent)
        thread_name = email_service.get_thread_name()
        self.assertEqual(thread_name, str(email_service.comment.parent).split(':')[0])

    def test_send_notification_to_followers_return_none(self):
        """return None if the thread has no followers"""
        followers = Follower.objects.filter_for_model_object(
            self.comment_obj.content_object).exclude(email=self.comment_obj.email)
        self.assertEqual(followers.count(), 0)
        result = self.email_service.send_notification_to_followers()
        self.assertIsNone(result)
        self.assertEqual(len(mail.outbox), 0)

    def test_send_notification_to_followers(self):
        Follower.objects.follow('any@test.com', 'test_user', self.comment_obj.content_object)
        followers = Follower.objects.filter_for_model_object(
            self.comment_obj.content_object).exclude(email=self.comment_obj.email)
        self.assertEqual(followers.count(), 1)

        self.email_service.send_notification_to_followers()
        self.email_service._email_thread.join()

        self.assertEqual(len(mail.outbox), 1)
        sent_email = mail.outbox[0]
        self.assertIsInstance(sent_email, EmailMultiAlternatives)
        username = self.comment_obj.get_username()
        thread_name = str(self.comment_obj.content_object)
        self.email_contents_test(sent_email.body, username, self.comment_obj.content)
        self.assertEqual(
            sent_email.subject,
            EmailInfo.NOTIFICATION_SUBJECT.format(username=username, thread_name=thread_name)
        )
        self.receivers = [followers.first().email]
        self.email_metadata_test(sent_email)



from unittest.mock import patch

from django.contrib.contenttypes.models import ContentType

from comment.conf import settings
from comment.models import Follower
from comment.tests.base import BaseCommentTest


class FollowerModelTest(BaseCommentTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment_test_follow = cls.create_comment(cls.content_object_1)
        cls.email = 't@t.com'
        cls.follower = Follower.objects.create(
            email=cls.email,
            username='test',
            content_object=cls.comment_test_follow
        )

    def test_can_create_entry(self):
        self.assertIsNotNone(self.follower)

    def test_string_value(self):
        self.assertEqual(str(self.follower), f'{str(self.comment_test_follow)} followed by {self.email}')
        self.assertEqual(repr(self.follower), f'{str(self.comment_test_follow)} followed by {self.email}')


class FollowerManagerTest(BaseCommentTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.manager = Follower.objects
        cls.follower_email = 'f1@t.com'
        cls.unfollower_email = 'uf@t.com'
        cls.comment_test_follow = cls.create_comment(cls.content_object_1)
        cls.comment_without_email = cls.create_comment(cls.content_object_1, user=cls.user_without_email)
        cls.follower = cls.manager.create(
            email=cls.follower_email,
            username='test',
            content_object=cls.comment_test_follow
        )

    def test_is_following(self):
        self.assertTrue(self.manager.is_following(self.follower_email, self.comment_test_follow))
        self.assertFalse(self.manager.is_following(self.unfollower_email, self.comment_test_follow))

    def test_follow_return_none_on_missing_email(self):
        self.assertIsNone(self.manager.follow('', 'username', self.comment_test_follow))

    def test_follow_return_none_if_email_is_already_follow(self):
        self.assertTrue(self.manager.is_following(self.follower_email, self.comment_test_follow))
        self.assertIsNone(self.manager.follow(self.follower_email, 'username', self.comment_test_follow))

    def test_follow_create_follower_instance(self):
        initial_count = self.manager.count()
        follower = self.manager.follow(self.unfollower_email, 'username', self.comment_test_follow)
        self.assertIsInstance(follower, self.manager.model)
        self.assertEqual(self.manager.count(), initial_count + 1)

    def test_unfollow_delete_follower_instance(self):
        initial_count = self.manager.count()
        self.assertTrue(self.manager.is_following(self.follower_email, self.comment_test_follow))
        self.manager.unfollow(self.follower_email, self.comment_test_follow)
        self.assertEqual(self.manager.count(), initial_count - 1)

    def test_toggle_follow_return_false_on_missing_email(self):
        email = None
        result = self.manager.toggle_follow(email=email, username='test', model_object=self.comment_test_follow)
        self.assertFalse(result)

    def test_toggle_follow_for_follower(self):
        """set the follower to unfollower and return false"""
        self.assertTrue(self.manager.is_following(self.follower_email, self.comment_test_follow))
        result = self.manager.toggle_follow(
            email=self.follower_email,
            username='test_user',
            model_object=self.comment_test_follow
        )
        self.assertFalse(result)
        self.assertFalse(self.manager.is_following(self.follower_email, self.comment_test_follow))

    def test_toggle_follow_for_unfollower(self):
        """set the unfollower to follower and return true"""
        self.assertFalse(self.manager.is_following(self.unfollower_email, self.comment_test_follow))
        result = self.manager.toggle_follow(
            email=self.unfollower_email,
            username='test_user',
            model_object=self.comment_test_follow
        )
        self.assertTrue(result)
        self.assertTrue(self.manager.is_following(self.unfollower_email, self.comment_test_follow))

    def test_follow_parent_thread_for_comment_no_email(self):
        self.assertFalse(self.comment_without_email.email)
        self.assertFalse(self.manager.is_following(self.comment_without_email.email, self.comment_without_email))
        self.manager.follow_parent_thread_for_comment(self.comment_without_email)
        self.assertFalse(self.manager.is_following(self.comment_without_email.email, self.comment_without_email))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_follow_parent_thread_for_comment_child_comment(self):
        child_comment = self.create_comment(self.content_object_1, user=self.user_2, parent=self.comment_without_email)

        # the parent (thread) will not be followed on creating child comment
        self.assertFalse(self.manager.is_following(child_comment.email, child_comment.content_object))
        # the parent comment (thread) is not followed yet
        self.assertFalse(self.manager.is_following(child_comment.email, self.comment_without_email))
        # child comment cannot be followed
        self.assertFalse(self.manager.is_following(child_comment.email, child_comment))

        self.manager.follow_parent_thread_for_comment(child_comment)

        # the parent (thread) will not be followed on creating child comment
        self.assertFalse(self.manager.is_following(child_comment.email, child_comment.content_object))
        # the parent is now followed
        self.assertTrue(self.manager.is_following(child_comment.email, self.comment_without_email))
        # child comment cannot be followed
        self.assertFalse(self.manager.is_following(child_comment.email, child_comment))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_follow_parent_thread_for_comment_parent_comment(self):
        parent_comment = self.create_comment(self.content_object_1, user=self.user_2)
        # the parent (thread) is not followed yet
        self.assertFalse(self.manager.is_following(parent_comment.email, parent_comment.content_object))
        # parent comment is not followed yet
        self.assertFalse(self.manager.is_following(parent_comment.email, parent_comment))

        self.manager.follow_parent_thread_for_comment(parent_comment)

        # the parent (thread) is now followed
        self.assertTrue(self.manager.is_following(parent_comment.email, parent_comment.content_object))
        # parent comment is now followed
        self.assertTrue(self.manager.is_following(parent_comment.email, parent_comment))

    def test_get_all_followers_for_model_object(self):
        followers = self.manager.filter_for_model_object(self.comment_test_follow)
        content_type = ContentType.objects.get_for_model(self.comment_test_follow)
        self.assertNotEqual(followers.count(), 0)
        self.assertEqual(
            list(followers),
            list(self.manager.filter(content_type=content_type, object_id=self.comment_test_follow.id))
        )

    def test_get_get_emails_for_model_object(self):
        emails = self.manager.get_emails_for_model_object(self.comment_test_follow)
        self.assertIn(self.comment_test_follow.email, emails)



from comment.messages import ExceptionError

try:
    from rest_framework.exceptions import APIException
except ModuleNotFoundError:
    APIException = Exception


class CommentBadRequest(APIException):
    status_code = 400
    default_detail = ExceptionError.BAD_REQUEST

    def __init__(self, detail=None, status_code=None):
        if status_code:
            self.status_code = status_code
        if not detail:
            detail = self.default_detail
        self.detail = detail



from django.utils.translation import gettext_lazy as _


class ErrorMessage:
    LOGIN_URL_MISSING = _('Comment App: LOGIN_URL is not in the settings')
    METHOD_NOT_IMPLEMENTED = _('Your {class_name} class has not defined a {method_name} method, which is required.')
    NON_AJAX_REQUEST = _('Only AJAX request are allowed')
    INVALID_ORDER_ARGUMENT = _((
        'Comment app: "{order}" is not a valid value for COMMENT_ORDER_BY. '
        'Please choose one among {allowed_orders}.'
        ))
    DUPLICATE_ORDER_VALUE = _((
        'Comment app: COMMENT_ORDER_BY should not have duplicated values '
        'Duplicated Values: {duplicates}. Please use one value only E.g. "{order}" or "-{order}".'
    ))
    WRAP_CONTENT_WORDS_NOT_INT = _('Comment App: settings var COMMENT_WRAP_CONTENT_WORDS must be an integer')


class ExceptionError:
    ERROR_TYPE = _('error')
    BAD_REQUEST = _('Bad Request')


class ContentTypeError:
    ID_NOT_INTEGER = _('{var_name} id must be an integer, {id} is NOT')
    APP_NAME_MISSING = _('app name must be provided')
    APP_NAME_INVALID = _('{app_name} is NOT a valid app name')
    MODEL_NAME_MISSING = _('model name must be provided')
    MODEL_NAME_INVALID = _('{model_name} is NOT a valid model name')
    MODEL_ID_MISSING = _('model id must be provided')
    MODEL_ID_INVALID = _('{model_id} is NOT a valid model id for the model {model_name}')
    PARENT_ID_INVALID = _('{parent_id} is NOT a valid id for a parent comment or '
                          'the parent comment does NOT belong to the provided model object')


class FlagError:
    SYSTEM_NOT_ENABLED = _('Flagging system must be enabled')
    NOT_FLAGGED_OBJECT = _('Object must be flagged!')
    STATE_INVALID = _('{state} is an invalid state')
    REASON_INVALID = _('{reason} is an invalid reason')
    INFO_MISSING = _('Please supply some information as the reason for flagging')
    ALREADY_FLAGGED_BY_USER = _('This comment is already flagged by this user ({user})')
    NOT_FLAGGED_BY_USER = _('This comment was not flagged by this user ({user})')
    REJECT_UNFLAGGED_COMMENT = _('This action cannot be applied on unflagged comments')
    RESOLVE_UNEDITED_COMMENT = _('The comment must be edited before resolving the flag')
    STATE_CHANGE_ERROR = _('Unable to change flag state at the moment!')


class ReactionError:
    TYPE_INVALID = _('Reaction must be an valid ReactionManager.RelationType. {reaction_type} is not')


class EmailError:
    EMAIL_INVALID = _('Enter a valid email address.')
    EMAIL_MISSING = _('Email is required for posting anonymous comments.')
    BROKEN_VERIFICATION_LINK = _('The link seems to be broken.')
    USED_VERIFICATION_LINK = _('The comment has already been verified.')


class FlagInfo:
    FLAGGED_SUCCESS = _('Comment flagged')
    UNFLAGGED_SUCCESS = _('Comment flag removed')


class ReactionInfo:
    UPDATED_SUCCESS = _('Your reaction has been updated successfully')


class EmailInfo:
    CONFIRMATION_SUBJECT = _('Comment Confirmation Request')
    CONFIRMATION_SENT = _('We have sent a verification link to your email.'
                          'The comment will be displayed after it is verified.')
    INPUT_PLACEHOLDER = _('email address, this will be used for verification.')
    INPUT_TITLE = _('email address, it will be used for verification.')
    NOTIFICATION_SUBJECT = _('{username} added comment to "{thread_name}"')
    LABEL = _('email')


class FlagState:
    UNFLAGGED = _('Unflagged')
    FLAGGED = _('Flagged')
    REJECTED = _('Flag rejected by the moderator')
    RESOLVED = _('Comment modified by the author')


class FollowError:
    EMAIL_REQUIRED = _('Email is required to subscribe {model_object}')
    SYSTEM_NOT_ENABLED = _('Subscribe system must be enabled')



import random
import string
from enum import IntEnum, unique
import hashlib

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.core import signing
from django.apps import apps

from comment.conf import settings
from comment.messages import ErrorMessage


@unique
class CommentFailReason(IntEnum):
    BAD = 1
    EXISTS = 2


def get_model_obj(app_name, model_name, model_id):
    content_type = ContentType.objects.get(app_label=app_name, model=model_name.lower())
    model_object = content_type.get_object_for_this_type(id=model_id)
    return model_object


def is_gravatar_enabled():
    return getattr(settings, 'COMMENT_USE_GRAVATAR')


def get_gravatar_img(email):
    if not is_gravatar_enabled() or not email:
        return settings.COMMENT_DEFAULT_PROFILE_PIC_LOC
    hashed_email = hashlib.md5(email.lower().encode('utf-8')).hexdigest()
    return f'https://www.gravatar.com/avatar/{hashed_email}'


def get_profile_content_type():
    profile_app_name = getattr(settings, 'PROFILE_APP_NAME', None)
    profile_model_name = getattr(settings, 'PROFILE_MODEL_NAME', None)
    if not profile_app_name or not profile_model_name:
        return None
    try:
        content_type = ContentType.objects.get(
            app_label=profile_app_name,
            model=profile_model_name.lower()
        )
    except ContentType.DoesNotExist:
        return None

    return content_type


def get_profile_instance(user):
    try:
        return getattr(user, settings.PROFILE_MODEL_NAME.lower(), None)
    except AttributeError:
        return None


def has_valid_profile():
    if getattr(settings, 'COMMENT_USE_GRAVATAR'):
        return True

    content_type = get_profile_content_type()
    if not content_type:
        return False
    profile_model = content_type.model_class()
    fields = profile_model._meta.get_fields()
    for field in fields:
        if hasattr(field, "upload_to"):
            return True
    return False


def is_comment_admin(user):
    if settings.COMMENT_FLAGS_ALLOWED:
        return user.groups.filter(name="comment_admin").exists() or (
            user.has_perm("comment.delete_flagged_comment")
            and user.has_perm("comment.delete_comment")
        )
    return False


def is_comment_moderator(user):
    if settings.COMMENT_FLAGS_ALLOWED:
        return user.groups.filter(name="comment_moderator").exists() or user.has_perm(
            "comment.delete_flagged_comment"
        )
    return False


def paginate_comments(comments, comments_per_page, current_page):
    paginator = Paginator(comments, comments_per_page)
    try:
        return paginator.page(current_page)
    except PageNotAnInteger:
        return paginator.page(1)
    except EmptyPage:
        return paginator.page(paginator.num_pages)


def get_request_data(request, item, api=False):
    value = request.GET.get(item) or request.POST.get(item)
    if not value and api:
        value = request.data.get(item)
    return value


def id_generator(prefix='', chars=string.ascii_lowercase, len_id=6, suffix=''):
    return prefix + ''.join(random.choice(chars) for _ in range(len_id)) + suffix


def get_comment_from_key(key):
    class TmpComment:
        is_valid = True
        why_invalid = None
        obj = None

    temp_comment = TmpComment()
    comment_model = apps.get_model('comment', 'Comment')
    try:
        comment_dict = signing.loads(str(key))
        model_name = comment_dict.pop('model_name')
        model_id = comment_dict.pop('model_id')
        app_name = comment_dict.pop('app_name')
        comment_dict.update(
            {
                'content_object': get_model_obj(app_name, model_name, model_id),
                'parent': comment_model.objects.get_parent_comment(comment_dict['parent'])
            }
        )
        temp_comment.obj = comment_model(**comment_dict)

    except (ValueError, KeyError, AttributeError, signing.BadSignature):
        temp_comment.is_valid = False
        temp_comment.why_invalid = CommentFailReason.BAD

    if temp_comment.is_valid and comment_model.objects.comment_exists(temp_comment.obj):
        temp_comment.is_valid = False
        temp_comment.why_invalid = CommentFailReason.EXISTS
        temp_comment.obj = None
    return temp_comment


def get_user_for_request(request):
    if request.user.is_authenticated:
        return request.user
    return None


def get_wrapped_words_number():
    words_number = 0
    if settings.COMMENT_WRAP_CONTENT_WORDS is None:
        return words_number
    try:
        words_number = int(settings.COMMENT_WRAP_CONTENT_WORDS)
    except (ValueError, TypeError):
        raise ImproperlyConfigured(ErrorMessage.WRAP_CONTENT_WORDS_NOT_INT)

    return words_number



from django.core.exceptions import ValidationError
from django.shortcuts import get_object_or_404
from django.views import View

from comment.models import Comment, Flag, FlagInstance
from comment.mixins import CanSetFlagMixin, CanEditFlagStateMixin, DABResponseData
from comment.responses import UTF8JsonResponse
from comment.messages import FlagInfo, FlagError


class SetFlag(CanSetFlagMixin, View, DABResponseData):
    comment = None

    def get_object(self):
        self.comment = get_object_or_404(Comment, pk=self.kwargs.get('pk'))
        return self.comment

    def post(self, request, *args, **kwargs):
        self.data = {
            'status': 1
        }
        flag = Flag.objects.get_for_comment(self.comment)

        try:
            if FlagInstance.objects.set_flag(request.user, flag, **request.POST.dict()):
                self.msg = FlagInfo.FLAGGED_SUCCESS
                self.data['flag'] = 1
            else:
                self.msg = FlagInfo.UNFLAGGED_SUCCESS

            self.data.update({'status': 0})
        except ValidationError as e:
            self.error = e.message
            self.status = 400

        return UTF8JsonResponse(self.json(), status=self.status)


class ChangeFlagState(CanEditFlagStateMixin, View, DABResponseData):
    comment = None

    def get_object(self):
        self.comment = get_object_or_404(Comment, pk=self.kwargs.get('pk'))
        return self.comment

    def post(self, request, *args, **kwargs):
        state = request.POST.get('state')
        try:
            self.comment.flag.toggle_state(state, request.user)
        except ValidationError:
            self.error = FlagError.STATE_CHANGE_ERROR
            self.status = 400

        self.data = {
            'state': self.comment.flag.state
        }
        return UTF8JsonResponse(self.json(), status=self.status)



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone

from comment.conf import settings
from comment.managers import FlagManager, FlagInstanceManager
from comment.models import Comment
from comment.messages import FlagError, FlagState


User = get_user_model()


class Flag(models.Model):
    UNFLAGGED = 1
    FLAGGED = 2
    REJECTED = 3
    RESOLVED = 4
    STATES_CHOICES = [
        (UNFLAGGED, FlagState.UNFLAGGED),
        (FLAGGED, FlagState.FLAGGED),
        (REJECTED, FlagState.REJECTED),
        (RESOLVED, FlagState.RESOLVED),
    ]

    comment = models.OneToOneField(Comment, on_delete=models.CASCADE)
    count = models.PositiveIntegerField(default=0)
    state = models.SmallIntegerField(choices=STATES_CHOICES, default=UNFLAGGED)
    moderator = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        null=True, blank=True,
        related_name='flags_moderated'
        )

    objects = FlagManager()

    def increase_count(self):
        """Increase flag count and save the model """
        self.refresh_from_db()
        field = 'count'
        self.count = models.F(field) + 1
        self.save(update_fields=[field])

    def decrease_count(self):
        """Decrease flag count and save the model """
        self.refresh_from_db()
        field = 'count'
        self.count = models.F(field) - 1
        self.save(update_fields=[field])

    @property
    def comment_author(self):
        return self.comment.user

    def get_verbose_state(self, state):
        state = self.get_clean_state(state)
        for item in self.STATES_CHOICES:
            if item[0] == state:
                return item[1]
        return None

    @property
    def is_flag_enabled(self):
        return bool(getattr(settings, 'COMMENT_FLAGS_ALLOWED', 0))

    def get_clean_state(self, state):
        err = ValidationError(FlagError.STATE_INVALID.format(state=state), code='invalid')
        try:
            state = int(state)
            if state not in [st[0] for st in self.STATES_CHOICES]:
                raise err
        except (ValueError, TypeError):
            raise err
        return state

    def toggle_state(self, state, moderator):
        state = self.get_clean_state(state)
        # toggle states occurs between rejected and resolved states only
        if state != self.REJECTED and state != self.RESOLVED:
            raise ValidationError(FlagError.STATE_INVALID.format(state=state), code='invalid')
        if self.state == state:
            self.state = self.FLAGGED
        else:
            self.state = state
        self.moderator = moderator
        self.save()

    def toggle_flagged_state(self):
        allowed_flags = getattr(settings, 'COMMENT_FLAGS_ALLOWED', 0)
        if not allowed_flags:
            return
        self.refresh_from_db()
        if self.count > allowed_flags and self.state not in [self.RESOLVED, self.REJECTED]:
            self.state = self.FLAGGED
            self.save()
        else:
            self.state = self.UNFLAGGED
            self.save()


class FlagInstance(models.Model):
    flag = models.ForeignKey(Flag, on_delete=models.CASCADE, related_name='flags')
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='flags')
    info = models.TextField(null=True, blank=True)
    date_flagged = models.DateTimeField(auto_now=timezone.now())

    objects = FlagInstanceManager()
    reason = models.SmallIntegerField(choices=objects.REASONS, default=objects.reason_values[0])

    class Meta:
        unique_together = ('flag', 'user')
        ordering = ('date_flagged',)



from django.dispatch import receiver
from django.db.models import signals

from comment.models import Comment, Flag, FlagInstance, Reaction, ReactionInstance, Follower
from comment.conf import settings


@receiver(signals.post_save, sender=Comment)
def add_initial_instances(sender, instance, created, raw, using, update_fields, **kwargs):
    if created:
        Reaction.objects.create(comment=instance)
        Flag.objects.create(comment=instance)
        if settings.COMMENT_ALLOW_SUBSCRIPTION:
            Follower.objects.follow_parent_thread_for_comment(comment=instance)


@receiver(signals.post_save, sender=FlagInstance)
def increase_count(sender, instance, created, raw, using, update_fields, **kwargs):
    if created:
        instance.flag.increase_count()
        instance.flag.toggle_flagged_state()


@receiver(signals.post_save, sender=ReactionInstance)
def add_count(sender, instance, created, raw, using, update_fields, **kwargs):
    if created:
        instance.reaction.increase_reaction_count(instance.reaction_type)



from unittest.mock import patch

from django.contrib.auth.models import AnonymousUser

from comment.forms import CommentForm
from comment.tests.base import BaseCommentTest, RequestFactory
from comment.conf import settings
from comment.messages import EmailInfo


class TestCommentForm(BaseCommentTest):
    def setUp(self):
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get('/')

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', False)
    def test_create_comment_form(self):
        self.request.user = self.user_1
        form = CommentForm(request=self.request)

        self.assertIsNone(form.fields.get('email', None))

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_anonymous_comment_form(self):
        self.request.user = AnonymousUser()
        form = CommentForm(request=self.request)
        field = 'email'
        email_field = form.fields.get(field, None)

        self.assertIsNotNone(email_field)
        self.assertEqual(email_field.label, EmailInfo.LABEL)
        self.assertTextTranslated(email_field.label)
        self.assertEqual(email_field.widget.input_type, field)
        self.assertDictEqual(email_field.widget.attrs, {
            'placeholder': EmailInfo.INPUT_PLACEHOLDER,
            'title': EmailInfo.INPUT_TITLE
        })

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_anonymous_comment_without_email(self):
        self.request.user = AnonymousUser()
        field = 'email'
        email = 'A@a.com '
        form = CommentForm(request=self.request, data={field: email})

        self.assertEqual(form.is_valid(), False)
        self.assertEqual(form.clean_email(), email.lower().strip())



from unittest.mock import patch

from django.core.exceptions import ValidationError

from comment.conf import settings
from comment.models import FlagInstance
from comment.tests.base import BaseCommentFlagTest


class FlagInstanceModelTest(BaseCommentFlagTest):
    def test_create_flag(self):
        data = self.flag_data
        comment = self.comment
        instance = self.create_flag_instance(self.user, comment, **data)
        self.assertIsNotNone(instance)
        comment.refresh_from_db()
        self.assertEqual(comment.flag.count, 1)


class FlagInstanceManagerTest(BaseCommentFlagTest):
    def test_clean_reason_for_invalid_value(self):
        data = self.flag_data.copy()
        data.update({'reason': -1})
        self.assertRaises(ValidationError, self.set_flag, self.user, self.comment, **data)

        data.update({'reason': 'abc'})
        self.assertRaises(ValidationError, self.set_flag, self.user, self.comment, **data)

    def test_clean_for_invalid_values(self):
        data = self.flag_data.copy()
        user = self.user
        comment = self.comment
        # info can't be blank with the last reason(something else)
        data.update({'reason': FlagInstance.objects.reason_values[-1]})
        self.assertRaises(ValidationError, self.set_flag, user, comment, **data)

        data.pop('reason')
        self.assertRaises(ValidationError, self.set_flag, user, comment, **data)

    def test_clean_ignores_info_for_all_reasons_except_last(self):
        data = self.flag_data.copy()
        info = 'Hi'
        data['info'] = info
        user = self.user
        comment = self.comment
        self.set_flag(user, comment, **data)
        instance = FlagInstance.objects.get(user=user, flag=comment.flag)

        self.assertIsNone(instance.info)

        new_comment = self.create_comment(self.content_object_1)
        data['reason'] = FlagInstance.objects.reason_values[-1]
        self.set_flag(user, new_comment, **data)
        instance = FlagInstance.objects.get(user=user, flag=new_comment.flag)

        self.assertEqual(instance.info, info)

    def test_set_flag_for_create(self):
        self.assertTrue(self.set_flag(self.user, self.comment, **self.flag_data))

    def test_set_flag_for_delete(self):
        self.assertFalse(self.set_flag(self.user_2, self.comment_2))

    def test_create_flag_twice(self):
        self.assertTrue(self.set_flag(self.user, self.comment, **self.flag_data))
        self.assertRaises(ValidationError, self.set_flag, self.user, self.comment, **self.flag_data)

    def test_un_flag_non_exist_flag(self):
        # user try to un-flag comment that wasn't flagged yet
        self.assertRaises(ValidationError, self.set_flag, self.user, self.comment)


class FlagModelTest(BaseCommentFlagTest):
    def test_flag_count(self):
        comment = self.comment
        self.assertEqual(comment.flag.count, 0)
        comment.flag.increase_count()
        comment.refresh_from_db()
        self.assertEqual(comment.flag.count, 1)
        comment.flag.decrease_count()
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.count, 0)

    def test_comment_author(self):
        comment = self.comment
        self.assertEqual(comment.user, comment.flag.comment_author)

    def test_is_flagged_enabled(self):
        flag = self.create_comment(self.content_object_1).flag
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1):
            self.assertIs(True, flag.is_flag_enabled)

        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0):
            self.assertIs(False, flag.is_flag_enabled)

    @patch('comment.models.flags.Flag.get_clean_state')
    def test_get_verbose_state(self, mocked_get_clean_state):
        flag = self.create_comment(self.content_object_1).flag
        mocked_get_clean_state.return_value = flag.FLAGGED
        self.assertEqual(flag.get_verbose_state(flag.FLAGGED), flag.STATES_CHOICES[flag.FLAGGED-1][1])
        mocked_get_clean_state.return_value = 100
        self.assertIsNone(flag.get_verbose_state(100))

    def test_get_clean_state(self):
        flag = self.create_comment(self.content_object_1).flag
        state = flag.get_clean_state(flag.FLAGGED)
        self.assertEqual(state, 2)

        # int not in existing states
        self.assertRaises(ValidationError, flag.get_clean_state, 100)

        # not int
        self.assertRaises(ValidationError, flag.get_clean_state, 'Not int')

        # None
        self.assertRaises(ValidationError, flag.get_clean_state, None)

    def test_toggle_state(self):
        flag = self.create_comment(self.content_object_1).flag
        self.assertIsNone(flag.moderator)
        self.assertEqual(flag.state, flag.UNFLAGGED)

        # toggle states occurs between rejected and resolved only
        self.assertRaises(ValidationError, flag.toggle_state, flag.FLAGGED, self.moderator)

        flag.toggle_state(flag.REJECTED, self.moderator)
        self.assertEqual(flag.state, flag.REJECTED)
        self.assertEqual(flag.moderator, self.moderator)

        # passing RESOLVED state value for the first time
        flag.toggle_state(flag.RESOLVED, self.moderator)
        self.assertEqual(flag.state, flag.RESOLVED)

        # passing RESOLVED state value for the second time
        flag.toggle_state(flag.RESOLVED, self.moderator)
        # state reset to FLAGGED
        self.assertEqual(flag.state, flag.FLAGGED)

    def test_toggle_flagged_state(self):
        comment = self.create_comment(self.content_object_1)
        flag = comment.flag
        flag.toggle_flagged_state()
        self.assertEqual(flag.state, flag.UNFLAGGED)

        # TODO split this test to be independent from the actual settings
        # current settings value COMMENT_FLAGS_ALLOWED = 2
        self.create_flag_instance(self.user_1, comment)
        self.create_flag_instance(self.user_2, comment)
        flag.refresh_from_db()
        self.assertEqual(flag.count, 2)

        # flagging is disabled => state won't change
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0):
            flag.toggle_flagged_state()
            self.assertEqual(flag.state, flag.UNFLAGGED)

        # flagging is enabled => state changes
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1):
            flag.toggle_flagged_state()
            self.assertEqual(flag.state, flag.FLAGGED)

        # increase allowed flags count => change the state to UNFLAGGED
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 10):
            self.assertEqual(flag.count, 2)
            flag.toggle_flagged_state()
            self.assertEqual(flag.state, flag.UNFLAGGED)



from django.contrib.contenttypes.models import ContentType
from django.db import models

from comment.conf import settings
from comment.utils import id_generator
from comment.validators import _validate_order


class CommentManager(models.Manager):

    def all_exclude_flagged(self):
        """Filter out comments that have been flagged"""
        allowed_flags = getattr(settings, 'COMMENT_FLAGS_ALLOWED', 0)
        show_flagged = getattr(settings, 'COMMENT_SHOW_FLAGGED', False)
        if not allowed_flags or show_flagged:
            return super().get_queryset()

        return super().get_queryset().exclude(flag__state__exact=2)

    @staticmethod
    def _filter_parents(qs, parent=None):
        return qs.filter(parent=parent).order_by(*_validate_order())

    def all_parents(self):
        qs = self.all_exclude_flagged()
        return self._filter_parents(qs)

    def all_comments_by_object(self, obj, include_flagged=False):
        content_type = ContentType.objects.get_for_model(obj.__class__)
        if include_flagged:
            return self.filter(content_type=content_type, object_id=obj.id)
        return self.all_exclude_flagged().filter(content_type=content_type, object_id=obj.id)

    def filter_parents_by_object(self, obj, include_flagged=False):
        if include_flagged:
            qs = self.all_comments_by_object(obj, include_flagged=True)
        else:
            qs = self.all_comments_by_object(obj)

        return self._filter_parents(qs)

    @staticmethod
    def generate_urlhash():
        return id_generator(
            prefix=settings.COMMENT_URL_PREFIX,
            len_id=settings.COMMENT_URL_ID_LENGTH,
            suffix=settings.COMMENT_URL_SUFFIX
            )

    def get_parent_comment(self, parent_id):
        parent_comment = None
        if parent_id or parent_id == '0':
            parent_qs = self.filter(id=parent_id)
            if parent_qs.exists():
                parent_comment = parent_qs.first()
        return parent_comment

    def comment_exists(self, comment):
        return self.model.objects.filter(email=comment.email, posted=comment.posted).count() > 0



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from rest_framework import status

from comment.conf import settings
from comment.mixins import AJAXRequiredMixin, BasePermission, ObjectLevelMixin
from comment.messages import ErrorMessage
from comment.tests.base import BaseCommentMixinTest, BaseCommentFlagTest


class AJAXMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.mixin = AJAXRequiredMixin()

    def test_non_ajax_request(self):
        request = self.factory.get('/')
        response = self.mixin.dispatch(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.content.decode('utf-8'), ErrorMessage.NON_AJAX_REQUEST)


class BasePermissionTest(BaseCommentMixinTest):
    def test_has_permission(self):
        request = self.factory.get('/')
        self.assertTrue(BasePermission().has_permission(request))

    def test_has_object_permission(self):
        request = self.factory.get('/')
        self.assertTrue(BasePermission().has_object_permission(request, 'object'))


class CanCreateMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.url = reverse('comment:create')
        self.client.force_login(self.user_1)

    def test_logged_in_user_permission(self):
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', False)
    def test_logged_out_user_permission(self, ):
        self.client.logout()
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 302)
        # user will be redirected to login
        self.assertEqual(response.url, settings.LOGIN_URL + '?next=/comment/create/')

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_permission_when_anonymous_comment_allowed(self):
        self.client.logout()
        self.data['email'] = 'test@test.come'
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)


class ObjectLevelMixinTest(BaseCommentMixinTest):
    def test_get_object_without_overriding(self):
        object_mixin = ObjectLevelMixin()
        self.assertRaises(ImproperlyConfigured, object_mixin.get_object)


class CanEditMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('comment:edit', kwargs={'pk': self.comment.id})

    def test_comment_owner_can_edit(self):
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

    def test_edit_comment_by_non_owner(self):
        self.assertNotEqual(self.comment.user.id, self.user_2.id)
        self.client.force_login(self.user_2)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

    def test_edit_comment_by_anonymous(self):
        """anonymous will be redirected to login page"""
        self.client.logout()
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, settings.LOGIN_URL + f'?next=/comment/edit/{self.comment.id}/')


class CanDeleteMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('comment:delete', kwargs={'pk': self.comment.id})

    def test_delete_comment_by_owner(self):
        self.assertEqual(self.comment.user.id, self.user_1.id)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

    def test_delete_comment_by_non_owner(self):
        self.assertNotEqual(self.comment.user.id, self.user_2.id)
        self.client.force_login(self.user_2)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

    def test_delete_comment_by_admin(self):
        self.assertNotEqual(self.comment.user.id, self.admin.id)
        self.client.force_login(self.admin)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

    def test_delete_comment_by_moderator(self):
        """moderator cannot delete comment unless it's flagged"""
        self.assertNotEqual(self.comment.user.id, self.moderator.id)
        self.client.force_login(self.moderator)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

    def test_delete_flagged_comment_by_moderator(self):
        """moderator cannot delete comment unless it's flagged"""
        self.assertNotEqual(self.comment.user.id, self.moderator.id)
        self.client.force_login(self.moderator)
        self.create_flag_instance(self.user_2, self.comment)
        self.create_flag_instance(self.admin, self.comment)
        self.create_flag_instance(self.moderator, self.comment)
        self.assertEqual(self.flags, 3)
        self.assertTrue(self.comment.is_flagged)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)


class BaseFlagMixinTest(BaseCommentFlagTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('comment:flag', kwargs={'pk': self.comment_2.id})

    @patch('comment.mixins.settings', COMMENT_FLAGS_ALLOWED=0)
    def test_flag_not_enabled_permission(self, _):
        """permission denied when flagging not enabled"""
        self.client.force_login(self.user_2)
        response = self.client.post(self.url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

    @patch('comment.mixins.settings', COMMENT_FLAGS_ALLOWED=2)
    def test_flag_enabled_permission(self, _):
        self.client.force_login(self.user_2)
        response = self.client.post(self.url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)


class CanSetFlagMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('comment:flag', kwargs={'pk': self.comment.id})

    def test_user_cannot_flag_their_own_comment(self):
        self.assertEqual(self.comment.user.id, self.user_1.id)
        self.client.force_login(self.user_1)
        response = self.client.post(self.url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)


class CanEditFlagStateMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('comment:flag-change-state', kwargs={'pk': self.comment.id})
        self.data = {'state': 3}

    def test_change_state_of_unflagged_comment(self):
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 400)

    def test_moderator_can_change_flag_state(self):
        self.create_flag_instance(self.user_2, self.comment)
        self.create_flag_instance(self.admin, self.comment)
        self.create_flag_instance(self.moderator, self.comment)
        self.comment.flag.refresh_from_db()
        self.assertEqual(self.flags, 3)
        self.assertTrue(self.comment.is_flagged)
        # normal user cannot change flag state
        self.client.force_login(self.user_2)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

        self.client.force_login(self.moderator)
        response = self.client.post(self.url, data=self.data, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)


class CanSubscribeMixinTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.toggle_follow_url = self.get_url(
            reverse('comment:toggle-subscription'), app_name='comment', model_name='comment', model_id=self.comment.id
        )

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_user_cannot_subscribe(self):
        self.client.force_login(self.user_1)
        self.assertFalse(settings.COMMENT_ALLOW_SUBSCRIPTION)
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 403)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_user_can_subscribe(self):
        self.client.force_login(self.user_1)
        self.assertTrue(settings.COMMENT_ALLOW_SUBSCRIPTION)
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)



from django.core.exceptions import ValidationError
from django.db import IntegrityError

from comment.models import ReactionInstance
from comment.tests.base import BaseCommentManagerTest, BaseCommentTest


class ReactionInstanceModelTest(BaseCommentManagerTest):
    def setUp(self):
        super().setUp()
        self.user = self.user_1
        self.comment = self.child_comment_1
        self.LIKE = ReactionInstance.ReactionType.LIKE.name
        self.DISLIKE = ReactionInstance.ReactionType.DISLIKE.name

    def test_user_can_create_reaction(self):
        """Test whether reaction instance can be created"""
        instance = self.create_reaction_instance(self.user, self.comment, self.LIKE)
        self.assertIsNotNone(instance)

    def test_unique_togetherness_of_user_and_reaction_type(self):
        """Test Integrity error is raised when one user is set to have more than 1 reaction type for the same comment"""
        self.create_reaction_instance(self.user, self.comment, self.LIKE)
        self.assertRaises(IntegrityError, self.create_reaction_instance, self.user, self.comment, self.DISLIKE)

    def test_comment_property_likes_increase_and_decrease(self):
        """Test decrease and increase on likes property with subsequent request."""
        comment = self.child_comment_2
        self.create_reaction_instance(self.user, comment, self.LIKE)
        comment.refresh_from_db()
        user = self.user_2
        self.create_reaction_instance(user, comment, self.LIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.likes, 2)

        self.set_reaction(user, comment, self.LIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.likes, 1)

    def test_comment_property_dislikes_increase_and_decrease(self):
        """Test decrease and increase on dislikes property with subsequent request."""
        comment = self.child_comment_3
        self.create_reaction_instance(self.user, comment, self.DISLIKE)
        comment.refresh_from_db()
        user = self.user_2
        self.create_reaction_instance(user, comment, self.DISLIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 2)

        # can't use create_reaction: one user can't create multiple reaction instances for a comment.
        self.set_reaction(user, comment, self.DISLIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 1)

    def test_set_reaction(self):
        """Test set reactions increments the likes and dislikes property appropriately for subsequent calls"""
        comment = self.comment
        user = self.user_2
        self.set_reaction(user, comment, self.DISLIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 1)
        self.assertEqual(comment.likes, 0)

        self.set_reaction(user, comment, self.DISLIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 0)
        self.assertEqual(comment.likes, 0)

        self.set_reaction(user, comment, self.LIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 0)
        self.assertEqual(comment.likes, 1)

        self.set_reaction(user, comment, self.DISLIKE)
        comment.refresh_from_db()
        self.assertEqual(comment.dislikes, 1)
        self.assertEqual(comment.likes, 0)

    def test_set_reaction_on_incorrect_reaction(self):
        """Test ValidationError is raised when incorrect reaction type is passed"""
        self.assertRaises(ValidationError, self.set_reaction, self.user_1, self.child_comment_5, 'likes')


class ReactionModelTest(BaseCommentTest):
    def setUp(self):
        super().setUp()
        self.comment_1 = self.create_comment(self.content_object_1)
        self.comment_2 = self.create_comment(self.content_object_1)

    def test_reaction_count(self):
        self.assertEqual(self.comment_1.reaction.likes, 0)
        self.assertEqual(self.comment_1.reaction.dislikes, 0)

        self.comment_1.reaction.increase_reaction_count(ReactionInstance.ReactionType.LIKE.value)
        self.comment_1.reaction.refresh_from_db()

        self.assertEqual(self.comment_1.reaction.likes, 1)

        self.comment_1.reaction.increase_reaction_count(ReactionInstance.ReactionType.DISLIKE.value)
        self.comment_1.reaction.refresh_from_db()

        self.assertEqual(self.comment_1.reaction.dislikes, 1)

        self.comment_1.reaction.decrease_reaction_count(ReactionInstance.ReactionType.LIKE.value)
        self.comment_1.reaction.refresh_from_db()

        self.assertEqual(self.comment_1.reaction.likes, 0)

        self.comment_1.reaction.decrease_reaction_count(ReactionInstance.ReactionType.DISLIKE.value)
        self.comment_1.reaction.refresh_from_db()

        self.assertEqual(self.comment_1.reaction.dislikes, 0)


class ReactionInstanceManagerTest(BaseCommentTest):
    def test_clean_reaction_type(self):
        like = ReactionInstance.ReactionType.LIKE
        # valid reaction type
        reaction_type = ReactionInstance.objects.clean_reaction_type(like.name)
        self.assertEqual(reaction_type, like.value)

        # invalid reaction type
        self.assertRaises(ValidationError, ReactionInstance.objects.clean_reaction_type, 1)

        # invalid reaction type
        self.assertRaises(ValidationError, ReactionInstance.objects.clean_reaction_type, 'likes')
